/*
 Dahdi layer for Quasar-mini E1 interface board driver

 Copyright (C) 2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2012 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/list.h>
#include <linux/types.h>
#include <linux/interrupt.h>

#include "dahdi/kernel.h"

#include "dver.h"
#include "quasar_conf.h"
#include "quasar.h"

static struct quasar_span *qspan_from_dspan (struct dahdi_span *span) {
        return container_of(span, struct quasar_span, span);
}

static int qdahdi_maint(struct dahdi_span *span, int cmd)
{
	int res = 0;
	struct quasar_span * espan = qspan_from_dspan(span);
	struct quasar_dahdi * qd = espan->pvt;
	struct quasar * pq = qd->pvt;

	res = QUASAR_HAL(maint(span, cmd));


	return res;
}

static int qdahdi_rbsbits(struct dahdi_chan *chan, int bits)
{
	struct quasar_span * pspan = chan->pvt;
	struct quasar_dahdi * pd = pspan->pvt;
	struct quasar * pq = pd->pvt;

	if (pspan->span.lineconfig & DAHDI_CONFIG_CCS)
			return 0;

	return QUASAR_HAL(rbsbits(chan, bits));
}

#ifdef HAVE_SPAN_MAINTQ
static int qdahdi_chanconfig(struct dahdi_chan *chan, int sigtype)
#else
static int qdahdi_chanconfig(struct file *file, struct dahdi_chan *chan, int sigtype)
#endif //HAVE_SPAN_MAINTQ
{
	//hmm... in E1 case, just nothing to do...
	return 0;
}

static int qdahdi_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
}

static int qdahdi_close(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "qdahdi_close\n");
	module_put(THIS_MODULE);
	return 0;
}

static int qdahdi_open(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "qdahdi_open\n");
	try_module_get(THIS_MODULE);
	return 0;
}

static int qdahdi_shutdown(struct dahdi_span *span)
{
	struct quasar_span *qspan = qspan_from_dspan(span);
	struct quasar_dahdi * pd = qspan->pvt;
	struct quasar * pq = pd->pvt;
	
	QUASAR_HAL(shutdown(span));

	return 0;
}



#ifdef HAVE_SPAN_MAINTQ
static int qdahdi_startup(struct dahdi_span *span)
#else
static int qdahdi_startup(struct file *file, struct dahdi_span *span)
#endif //HAVE_SPAN_MAINTQ
{
	struct quasar_span *qspan = qspan_from_dspan(span);
    struct quasar_dahdi *pd = qspan->pvt;
	struct quasar * pq = pd->pvt;

	return QUASAR_HAL(startup(span));
}

#ifdef HAVE_SPAN_MAINTQ
static int qdahdi_spanconfig(struct dahdi_span *span, struct dahdi_lineconfig *lc)
#else
static int qdahdi_spanconfig(struct file *file, struct dahdi_span *span, struct dahdi_lineconfig *lc)
#endif //HAVE_SPAN_MAINTQ
{
	struct quasar_span *qspan = qspan_from_dspan(span);

	module_dprintk(KERN_DEBUG, "qdahdi_spanconfig: Configuring span %d, span=%p\n", span->spanno, span);
	span->lineconfig = lc->lineconfig;
	span->txlevel = 0;
	span->rxlevel = 0;
	qspan->syncprior = lc->sync; 	//lineconfig.sync contain the priority of the sysnc source 1-highest 1, 0 - not use this span for the sync at all.
	if (lc->lbo) {
		qspan->unframed = 1;
		module_printk(KERN_INFO, "Configuring span %d in unframed mode.\n", span->spanno);
	}
	
	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & DAHDI_FLAG_RUNNING)
	{
		module_dprintk(KERN_DEBUG, "Starting span!\n");
#ifdef HAVE_SPAN_MAINTQ
		return qdahdi_startup(span);
#else
		return qdahdi_startup(file, span);
#endif //HAVE_SPAN_MAINTQ
	}
	module_dprintk(KERN_DEBUG, "Done with spanconfig!\n");
	return 0;
}
void qdahdi_proc(struct quasar_dahdi * pd);

#ifdef QUASAR_TASKLET
static void quasar_tasklet(unsigned long data)
{
	struct quasar_dahdi *pd = (struct quasar_dahdi *)data;
	pd->taskletrun++;
	if (pd->taskletpending) {
		pd->taskletexec++;
		qdahdi_proc(pd);
	}
	pd->taskletpending = 0;
}
#endif


int qdahdi_dacs(struct dahdi_chan *dst, struct dahdi_chan *src);

#ifdef HAVE_SPAN_OPS
static const struct dahdi_span_ops qdahdi_span_ops = {
	.owner = THIS_MODULE,
	.spanconfig = qdahdi_spanconfig,
	.chanconfig = qdahdi_chanconfig,
	.startup = qdahdi_startup,
	.shutdown = qdahdi_shutdown,
	.rbsbits = qdahdi_rbsbits,
	.maint = qdahdi_maint,
	.open = qdahdi_open,
	.close = qdahdi_close,
	.ioctl = qdahdi_ioctl,
	.dacs = qdahdi_dacs,
};
void qdahdi_setops (struct dahdi_span *span) {
        span->ops = &qdahdi_span_ops;
}
#else //HAVE_SPAN_OPS
void qdahdi_setops (struct dahdi_span *span) {
	    span->spanconfig = qdahdi_spanconfig;
	    span->chanconfig = qdahdi_chanconfig;
	    span->startup = qdahdi_startup;
	    span->shutdown = qdahdi_shutdown;
	    span->rbsbits = qdahdi_rbsbits;
	    span->maint = qdahdi_maint;
	    span->open = qdahdi_open;
	    span->close  = qdahdi_close;
	    span->ioctl = qdahdi_ioctl;
	    span->dacs = qdahdi_dacs;
#ifdef HAVE_SPAN_OWNER
        span->owner = THIS_MODULE;
#endif //HAVE_SPAN_OWNER
}
#endif //HAVE_SPAN_OPS

#define DMA_BUFFER_SIZE 4096

int qdahdi_register(struct quasar * pq)
{
    struct pci_dev *pdev = pq->pci_device;
    struct quasar_dahdi * pd;
    int spanid, chanid;

    pq->priv_data = kmalloc(sizeof(struct quasar_dahdi), GFP_KERNEL);
    if (pq->priv_data == NULL) 
	return -ENOMEM;

    memset(pq->priv_data,0x0,sizeof(struct quasar_dahdi));
    pd = (struct quasar_dahdi *) pq->priv_data;
    pd->pvt = pq;
    
    pd->writechunk  = (unsigned int *)pci_alloc_consistent(pdev, 2 * DMA_BUFFER_SIZE, &pd->writedma);
    if (pd->writechunk == NULL)
    {
	    module_printk(KERN_ERR, "Unable to allocate DMA buffer\n");
	    kfree(pq->priv_data);
	    return -ENOMEM;
    }

    pd->readchunk = pd->writechunk + (DMA_BUFFER_SIZE / 4); //double-word address;
    pd->readdma = pd->writedma + DMA_BUFFER_SIZE; //physical byte address
	pq->readdma = pd->readdma;
    pq->writedma = pd->writedma;

    pd->num_spans = pq->e1_amount + 1;
    if (pd->num_spans == 2) 
	pd->qn_span = 2; 
    else 
	pd->qn_span = 4;

    /* Initialize Write/Buffers to all blank data */
    memset((void *)pd->writechunk,0x00, DMA_BUFFER_SIZE);
    memset((void *)pd->readchunk,0xff, DMA_BUFFER_SIZE);

    pci_set_master(pdev);

    //allocate memory for the spans
    for (spanid = 0; spanid < pd->num_spans; spanid++)
    {
	    pd->spans[spanid] = kmalloc(sizeof(struct quasar_span), GFP_KERNEL);
	    if (pd->spans[spanid])
	    {
		    memset(pd->spans[spanid], 0, sizeof(struct quasar_span));
	    }
	    else
	    {   //free already allocated
		    goto unreg_all;
	    }
    }
#ifdef HAVE_DAHDI_DEVICE
            pq->ddev = dahdi_create_device();
#endif //HAVE_DAHDI_DEVICE

    //initialize spans
    for (spanid = 0; spanid < pd->num_spans; spanid++) {
    	struct quasar_span *qspan = pd->spans[spanid];
		struct dahdi_span *span = &(qspan->span);

	    qspan->spannumber=spanid;
	    qspan->pvt = pd;
	    span->channels = pq->channels;
#ifdef HAVE_SPANTYPES
	    span->spantype = SPANTYPE_DIGITAL_E1;
#else
	    span->spantype = "E1";
#endif //HAVE_SPANTYPES
	    sprintf(span->name,
		    "Quasar_%d_%d", pq->dev_number,spanid+1);
	    sprintf(span->desc,
		    "Parabel Quasar %d Span %d",pq->dev_number, spanid+1);
#ifndef HAVE_DAHDI_DEVICE
	    span->manufacturer = "Parabel ltd.";
	    snprintf(span->location,
		     sizeof(span->location) - 1,
		     "PCI Bus %02d Slot %02d", pdev->bus->number,
		     PCI_SLOT(pdev->devfn) + 1);
	    snprintf(span->devicetype,
		     sizeof(span->devicetype) - 1,
		     "Quasar-mini PCI, hw:0x%X",
		     pq->bar0[BAR0_HW_FW_REV]);
#endif //HAVE_DAHDI_DEVICE

	    span->chans = pd->spans[spanid]->chans;
	    span->flags = DAHDI_FLAG_RBS;
	    span->linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS | DAHDI_CONFIG_CRC4 | DAHDI_CONFIG_AMI;  //which modes I can work
	    span->deflaw = DAHDI_LAW_ALAW;

	    qdahdi_setops (&pd->spans[spanid]->span);

#ifdef HAVE_SPAN_MAINTQ
	    init_waitqueue_head(&(pd->spans[spanid]->span.maintq));
#endif //HAVE_SPAN_MAINQ

	    for (chanid=0; chanid < pd->spans[spanid]->span.channels; chanid++)
	    {
	    	struct dahdi_chan *chan;
            chan = pd->spans[spanid]->chans[chanid] = &pd->spans[spanid]->_chans[chanid];
		    sprintf(chan->name, "Quasar_%d_%d_%d", pq->dev_number, spanid + 1, chanid + 1);
		    chan->sigcap = DAHDI_SIG_EM | DAHDI_SIG_CLEAR | DAHDI_SIG_EM_E1 |
			    DAHDI_SIG_FXSLS | DAHDI_SIG_FXSGS |
			    DAHDI_SIG_FXSKS | DAHDI_SIG_FXOLS |
			    DAHDI_SIG_FXOGS | DAHDI_SIG_FXOKS |
			    DAHDI_SIG_CAS | DAHDI_SIG_SF | DAHDI_SIG_DACS_RBS;
		    chan->pvt = pd->spans[spanid];
		    chan->chanpos = chanid + 1; //timeslot number
	    }
#ifdef HAVE_DAHDI_DEVICE
	    list_add_tail(&pd->spans[spanid]->span.device_node, &pq->ddev->spans);
#else
	    if (dahdi_register(span, 0))
	    {
	    	int j;
		    module_printk(KERN_ERR, "Unable to register span Quasar_%d_%d with dahdi!\n", pq->dev_number, spanid+1);
		    for (j=(spanid-1);j >=0; j--)
		    {
			    module_dprintk(KERN_DEBUG, "Unregister span %d\n", j);
			    dahdi_unregister(&pd->spans[spanid]->span);
		    }
		    goto unreg_all;
	    }
#endif //HAVE_DAHDI_DEVICE

    }
#ifdef HAVE_DAHDI_DEVICE
    pq->ddev->manufacturer = "Parabel ltd.";
    pq->ddev->devicetype = "Parabel Quasar E1";
    pq->ddev->location = kasprintf(GFP_KERNEL, "PCI Bus %02d Slot %02d",
	pdev->bus->number, PCI_SLOT(pdev->devfn) + 1);

    if (dahdi_register_device(pq->ddev, &pq->pci_device->dev)) {
	    module_printk(KERN_ERR, "Quasar_%d, Failed to register with DAHDI.\n", pq->dev_number);
	    goto unreg_all;
    }
#endif //HAVE_DAHDI_DEVICE

#ifdef QUASAR_TASKLET
	tasklet_init(&pd->quasar_tlet, quasar_tasklet, (unsigned long)pd);
#endif

     pd->registered = 1;
    return 0;
unreg_all:
	module_dprintk(KERN_DEBUG, "qdahdi_register: Free resources!\n");
	for (spanid = 0; spanid < pd->num_spans; spanid++) {
		if (pd->spans[spanid]) {
			kfree(pd->spans[spanid]);
			pd->spans[spanid] = NULL;
		}
	}
	pci_free_consistent(pdev, 2 * DMA_BUFFER_SIZE, (void *) pd->writechunk, pd->writedma);
	kfree(pq->priv_data);
	return -ENOMEM;
}
int qdahdi_unregister(struct quasar * pq)
{
	struct pci_dev *pdev = pq->pci_device;
	struct quasar_dahdi * pd;
	int j;
	pd = (struct quasar_dahdi *) pq->priv_data;
	if(pd)
	{
		if (pd->registered)
		{
#ifdef HAVE_DAHDI_DEVICE
		    kfree(pq->ddev->location);
		    dahdi_unregister_device(pq->ddev);
		    dahdi_free_device(pq->ddev);
#endif //HAVE_DAHDI_DEVICE
			//detach from dahdi
			for(j=0;j<pd->num_spans;j++)
			{
#ifndef HAVE_DAHDI_DEVICE
				dahdi_unregister(&pd->spans[j]->span);
#endif //HAVE_DAHDI_DEVICE
				kfree(pd->spans[j]);
			}
			pci_free_consistent(pdev, 2 * DMA_BUFFER_SIZE, (void *) pd->writechunk, pd->writedma);
			kfree(pq->priv_data);
			//deinitialize hardware

			QUASAR_HAL(stop_dma(pq));
		}
	}
    return 0;
}

void qdahdi_proc(struct quasar_dahdi * pd)
{
	int spanid, chanid;

    for(spanid=0; spanid < pd->num_spans; spanid++) {
    	struct quasar_span *qspan = pd->spans[spanid];

		if( qspan->span.flags &  DAHDI_FLAG_RUNNING)
		{
	    	/*Echo cancelation*/
			for(chanid=0; chanid < qspan->span.channels; chanid++) {
				dahdi_ec_chunk(qspan->span.chans[chanid],
					   	qspan->span.chans[chanid]->readchunk,
				    	qspan->ec_chunk2[chanid]);
				memcpy(qspan->ec_chunk2[chanid], qspan->ec_chunk1[chanid],
			    	   DAHDI_CHUNKSIZE);
				memcpy(qspan->ec_chunk1[chanid], qspan->span.chans[chanid]->writechunk,
			    	   DAHDI_CHUNKSIZE);
			}
			dahdi_receive(&qspan->span);
		}
    }
    for(spanid=0; spanid < pd->num_spans; spanid++)
    {
	    if( pd->spans[spanid]->span.flags &  DAHDI_FLAG_RUNNING)
	    {
		    dahdi_transmit(&pd->spans[spanid]->span);
	    }
    }
}

int qdahdi_delivery(struct quasar * pq) {
	struct quasar_dahdi *pd = pq->priv_data;

#ifdef QUASAR_TASKLET
	if (!pd->taskletpending) {
		pd->taskletpending = 1;
		pd->taskletsched++;
		tasklet_schedule(&pd->quasar_tlet);
	} else {
		module_dprintk(KERN_DEBUG, "Miss tasklet. Shed:%d Exec:%d Run:%d, TX_Err: %d\n",
			       pd->taskletsched, pd->taskletexec, pd->taskletrun, pd->txerrors++);
		pd->txerrors++;
	}
#endif

	QUASAR_HAL(copy_chunk(pq));

	QUASAR_HAL(rbs_delivery(pq));

#ifndef QUASAR_TASKLET
	qdahdi_proc(pd);
#endif
	++pd->intcount;
	if (0 == (pd->intcount & 127)) {
    	QUASAR_HAL(check_alarm (pd));
		QUASAR_HAL(quasar_find_master (pd));
		QUASAR_HAL(clearstat (pd));
	}

	return 0;
}


int qdahdi_dacs (struct dahdi_chan *dst, struct dahdi_chan *src) 
{
	struct quasar_span * dst_span = dst->pvt;
	struct quasar_dahdi * pd = dst_span->pvt;
	struct quasar * pq = pd->pvt;

	return QUASAR_HAL(dacs(dst,src));
}
