/*
 Quasar-mini interface board Linux driver.

 Copyright (C) 2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2011 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/device.h>
#if defined(HAVE_UNLOCKED_IOCTL) && defined(CONFIG_BKL)                                                                                                         
# include <linux/smp_lock.h> /* lock_kernel */
#endif
#include <linux/version.h>

#include "dahdi/kernel.h"

#include "kal.h"
#include "version.h"
#include "quasar_conf.h"
#include "quasar.h"
#include "pmc_e1_if.h"

#define DRV_NAME "quasar"
#define WIN_SIZE 0x10000 /* PCI window size */

static dev_t dev_file_glob; //descriptop for devfs and sysfs
static unsigned char device_number = 0;
static unsigned int cacheline = 0x20;
static unsigned int cacheline_bridge = 0x20;
static unsigned int latency = 0;
static unsigned int noec = 0;
static unsigned int ts32 = 0;
static unsigned int pcie_prefetch = 0;

static int irqflags = IRQF_SHARED | IRQF_DISABLED;
LIST_HEAD(dev_list_head);

static struct pci_device_id quasar_tbl[] = {
	{ PCI_VENDOR_ID_ALTERA, PCI_DEVICE_ID_QUASAR2012, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ PCI_VENDOR_ID_ALTERA, PCI_DEVICE_ID_QUASAR8, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ PCI_VENDOR_ID_ALTERA, PCI_DEVICE_ID_QUASARM, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ PCI_VENDOR_ID_ALTERA, PCI_DEVICE_ID_QUASARMICRO, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, quasar_tbl);

#ifdef PMC_E1_UDEV
static struct class *pmc_e1_class = NULL;
#else
# define quasar_sysfs_create(a)
# define quasar_sysfs_delete(a)
#endif //PMC_E1_UDEV

extern int qdahdi_register(struct quasar *); 
extern int qdahdi_delivery(struct quasar *); 
extern int qdahdi_unregister(struct quasar *);

extern struct quasar_hal_ops quasarm_hal_ops;
extern struct quasar_hal_ops quasar8_hal_ops;
extern struct quasar_hal_ops quasar2012_hal_ops;

static int pmc_e1_ioctl(struct inode * inode_p, struct file * file_p, unsigned int cmd, unsigned long arg)
{
    int r = 0;
    struct quasar * pq = 
	    (struct quasar *) file_p->private_data;

	r= QUASAR_HAL(ioctl(inode_p, file_p, cmd, arg));

    return r;
}

#ifdef HAVE_UNLOCKED_IOCTL
#  if !defined(CONFIG_BKL)
#    define lock_kernel();
#    define unlock_kernel();
#  endif //CONFIG_BKL
static long pmc_e1_ioctl_unlocked(struct file *file, unsigned int cmd, unsigned long data)
{
	long res;

	lock_kernel();
	res = pmc_e1_ioctl(NULL, file, cmd, data);
	unlock_kernel();

	return res;
}
#endif //HAVE_UNLOCKED_IOCTL

static int pmc_e1_open(struct inode * inode_p, struct file * file_p)
{
	file_p->private_data = container_of(inode_p->i_cdev, struct quasar, dev_file);
	return 0;
}

static int pmc_e1_release(struct inode * inode_p, struct file * file_p)
{
	return 0;
}

static struct file_operations pmc_e1_fops  = {
    owner:		THIS_MODULE,
    open:		pmc_e1_open,
#ifdef HAVE_UNLOCKED_IOCTL
	unlocked_ioctl:	pmc_e1_ioctl_unlocked,
#else
	ioctl:		pmc_e1_ioctl,
#endif
    release:            pmc_e1_release,
};

/*========== SYSFS code ==========*/
#ifdef PMC_E1_UDEV
struct ptmc_attribute {
    struct device_attribute attr;
    unsigned long p[3];
};
#define toptattr(a) ((struct ptmc_attribute *)(a))
#define todeattr(a) ((struct device_attribute *)(a))

#define PTMC_ATTR_RO(name) __ATTR(name, S_IRUGO, _pmc_show, NULL)
ssize_t _pmc_show(struct device *cls, struct device_attribute *attr,
		      char *buff)
{
    struct quasar * pq;
    unsigned long *p = toptattr(attr)->p;

    pq = (struct quasar *) dev_get_drvdata(cls);
    return sprintf(buff,"%ld\n", (pq->bar0[p[0]] >> p[1]) & p[2]);
}

ssize_t pmc_show(struct device *cls, struct device_attribute *attr, char *buff)
{
    int count=0,i;
    unsigned int tmp;
    struct quasar * pq;
    pq = (struct quasar *) dev_get_drvdata(cls);

    count+=sprintf(buff,"ChN:");
    for(i=0; i <= pq->e1_amount; i++)
	count+=sprintf(buff+count," %d",i);

    count+=sprintf(buff+count,"\nLOS:");
    for(i=0; i<=pq->e1_amount; i++)
    {
	tmp = pq->bar1[(0x40 * i) + FRS_ADDR_OFFSET];
	count+=sprintf(buff+count," %d", tmp & FRS_LOS ? 1 : 0);
    }

    spin_lock(&pq->dev_lock);
    count+=sprintf(buff+count,"\nLOF:");
    for(i=0; i<= pq->e1_amount; i++)
    {
	tmp = pq->bar1[(0x40 * i) + FRS_ADDR_OFFSET];
	count+=sprintf(buff+count," %d", tmp & FRS_RED ? 1 : 0);
    }
    count+=sprintf(buff+count,"\nLOM:");
    for(i=0; i <= pq->e1_amount;i++)
    {
	tmp = pq->bar1[(0x40 * i) + FRS_ADDR_OFFSET];
	count+=sprintf(buff+count," %d", !(tmp & FRS_MFRAME_SYNC) ? 1 : 0);
    }
    spin_unlock(&pq->dev_lock);

    count+=sprintf(buff+count,"\nWatchdog: ");
    if (pq->wd) count+=sprintf(buff+count,"present");
    else count+=sprintf(buff+count,"no");

    count+=sprintf(buff+count,"\nHW echo canceller: ");
    if (pq->ec) count+=sprintf(buff+count,"present");
    else count+=sprintf(buff+count,"no");

    count+=sprintf(buff+count,"\n");
    return count;
}
ssize_t wd_handle(struct device *cls, struct device_attribute *attr, char *buff)
{
    struct quasar * pq;
    pq = (struct quasar *) dev_get_drvdata(cls);

    if (pq->wd) {
		pq->bar0[BAR0_CTRL] |=  CTRL_WDOG;
		pq->bar0[BAR0_CTRL] &= ~CTRL_WDOG;
    }
    
    return 0;
}

static struct ptmc_attribute device_attrs[] = {
    {PTMC_ATTR_RO(dma_stat),    {0x09,8, 0xFF}},
    {PTMC_ATTR_RO(diff_cnt),    {0x09,16,0xFF}},
    {PTMC_ATTR_RO(dma_overflow),{0x09,24,0xFF}},
    {PTMC_ATTR_RO(pci_load_cur),{0x20,0, 0xFFFF}},
    {PTMC_ATTR_RO(pci_load), 	{0x20,16,0xFFFF}},
    {PTMC_ATTR_RO(pmc_load_max),{0x21,0, 0xFFFF}},
    {PTMC_ATTR_RO(pmc_load), 	{0x21,16,0xFFFF}},
    {__ATTR(info, S_IRUGO, pmc_show, NULL), {0,}},
    {__ATTR(wdog, S_IRUGO, wd_handle, NULL), {0,}},
};

static int quasar_sysfs_create (struct quasar * pq) {
    int i, error = 0;
    struct device *sysfs_device = NULL;

    pq->pmc_class_dev = sysfs_device =
			QUASAR_DEVICE_CREATE (pmc_e1_class, QUASAR_PARENT(pq),
		       MKDEV(MAJOR(dev_file_glob), pq->dev_number),
		       pq->name);
    if (IS_ERR(pq->pmc_class_dev)) {
        module_printk (KERN_WARNING, "Failed to device_create.\n");
	return 1;
    }
    dev_set_drvdata(pq->pmc_class_dev, (void *) pq);

    for (i = 0; i < ARRAY_SIZE(device_attrs); i++) {
	error = device_create_file(sysfs_device, &(device_attrs[i].attr));
	if (error)
	    break;
    }

    if (error) {
	while (--i >= 0)
	    device_remove_file(sysfs_device, todeattr(&device_attrs[i]));
    }

    return 0;
}

static void quasar_sysfs_delete (struct quasar * pq) {
    int i;
    struct device *sysfs_device = pq->pmc_class_dev;

    if(IS_ERR(pq->pmc_class_dev)) return;

    for (i = 0; i < ARRAY_SIZE(device_attrs); i++)
	device_remove_file(sysfs_device, todeattr(&device_attrs[i]));

    device_destroy(pmc_e1_class, MKDEV(MAJOR(dev_file_glob), pq->dev_number));
}
#endif //PMC_E1_UDEV

/*========== Init code & IRQ handler ==========*/
QUASAR_IRQ_HANDLER(quasar_interrupt_hr)
{
	unsigned int flags=0;
	struct quasar * pq = (struct quasar *) pr_data;

	if(pq == NULL ) return IRQ_NONE;
	flags = pq->bar0[BAR0_IFR];
    if (flags & 0x20000000) //check DMA interrupt;
	{
		pq->bar0[BAR0_IFR]=0x0;
	    qdahdi_delivery(pq);
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}

static int set_cacheline(struct pci_dev *pdev, u8 val)
{
	u8 cache_tmp;

	pci_write_config_byte(pdev, PCI_CACHE_LINE_SIZE, val);
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &cache_tmp);
	if (val == cache_tmp)
		return 0;

	module_printk(KERN_ERR, "Cache line size of %d DWORDs is not "
		"supported (%x:%x)\n", val, pdev->vendor, pdev->device);
	return -EINVAL;
}

/* Write 'val' to Main Control Register at address 'addr'. */
static int plx_write_mcr(struct pci_dev *pdev, u32 addr, u32 val)
{
	addr &= 0xFFF;	/* Remove offset '0x1000' */
	pci_write_config_dword(pdev, 0x84, addr);	/* Index */
	pci_write_config_dword(pdev, 0x88, val);	/* Data */

	return 0; /* Should I read & compare ? */
}

/* Read Main Control Register at address 'addr'. */
static u32 plx_read_mcr(struct pci_dev *pdev, u32 addr)
{
	u32 tmp_u32;

	addr &= 0xFFF;	/* Remove offset '0x1000' */
	pci_write_config_dword(pdev, 0x84, addr);	/* Index */
	pci_read_config_dword(pdev, 0x88, &tmp_u32);/* Data */

	return tmp_u32;
}

int tune_pci (struct pci_dev *pdev)
{
	struct pci_dev *bridge;
	u8 lat_timer, tmp_u8;
	u32 tmp_u32;

	if (QUASARMICRO(pdev))
		return 0;

 	module_printk (KERN_INFO, "Setting Cache Line Size to %d\n", cacheline);
	if (set_cacheline (pdev, cacheline))
            return -EINVAL;

	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &lat_timer);
	if (lat_timer < 40 || latency) {
		lat_timer = latency ? latency : 40;
		module_printk (KERN_INFO, "Setting latency timer to %d\n", lat_timer);
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER, lat_timer);
		pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &tmp_u8);
		if (tmp_u8 != lat_timer) {
			module_printk(KERN_ERR, "Latency timer of %d not supported (%x:%x)\n",
				lat_timer, pdev->vendor, pdev->device);
			return -EINVAL;
		}
	}


	if (!pdev->bus || !pdev->bus->self || !pdev->bus->self)
		return 0;

	bridge = pdev->bus->self;

	/* Configuring PCIe bridge (PLX 8112, PLX 8112), if any. */
	if ( 0x10b5 != bridge->vendor ||
		(0x8111 != bridge->device && 0x8112 != bridge->device))
		return 0;

	/* Hade a machine, who set PLX 811x "cache line size" to illegal value (>20) */
	module_printk (KERN_INFO, "Setting PCIe bridge Cache Line Size to %d\n",
		cacheline_bridge);
	if (set_cacheline (bridge, cacheline_bridge))
		return -EINVAL;

	/* Had a machine, who disabled PCI in PLX. */
	/* Have no idea who is responsible - BIOS or Linux, but ... Enabling it... */
	plx_write_mcr(bridge, 0x0, 0x33); /* PCI & PCIe enable */

	pci_read_config_dword(bridge, 0x48, &tmp_u32);
	tmp_u32 &= ~(1 | 1 << 11);
	if (pcie_prefetch)
		tmp_u32 |= (1 | 1 << 11);	/* enable blind prefetching & beacon generation */
	pci_write_config_dword(bridge, 0x48, tmp_u32);

	tmp_u32 = plx_read_mcr(bridge, 0x0c);
	tmp_u32 &= ~(7 << 27);
	if (pcie_prefetch)
		tmp_u32 |= 2 << 27;    /* set prefetch size to 128 bytes */
	plx_write_mcr(bridge, 0x0c, tmp_u32); /* PCI & PCIe enable */

	return 0;
}

/*
Allocates memory for quasar structure, returns a pointer
to allocated strunture, NULL if allocation insucessful.	*/
static struct quasar * quasar_alloc(void)
{
	struct quasar * pq;
	//create structure for pmc_device
	pq = kmalloc(sizeof(* pq), GFP_KERNEL);
	if(pq)
	{
		memset(pq,0, sizeof(*pq));
	}
	return pq;
}

static int quasar_init_one (struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct quasar * pq;
	int err;

	if (pci_enable_device(pdev))
		return -EIO;  //last
	if (device_number > (PMC_E18_NR-1))
		return -EBUSY;

	err = tune_pci (pdev);
	if (err)
	    return err;

	pq = quasar_alloc();
	if (!pq)
	{
		module_printk(KERN_ERR, "Failed to allocate device.\n");
		return -ENOMEM;
	}
    pq->dev_number = device_number;
    sprintf(pq->name,"quasar%d", device_number);

	//bind with pci device;
	pq->pci_device = pdev;
	//add to list of controllers, I don't remember why should I use this list, just part of old code from old project.
	list_add_tail(&pq->dev_node, &dev_list_head);
      // printk("pciaddr = 0x%lx, irq = %d, ptmc_dev = %p \n", pci_resource_start(pdev,0), pdev->irq, pq);
	if (request_mem_region (pci_resource_start(pdev,0), WIN_SIZE, pq->name) == NULL)
	{
	    module_printk (KERN_ERR, "I/O resource 0x%x @ 0x%llx busy\n",
		    WIN_SIZE, (u64) pci_resource_start(pdev,0));
	    goto fail_req_bar0;
	}
	if (request_mem_region (pci_resource_start (pdev, 1), WIN_SIZE, pq->name) == NULL)
	{
	    module_printk (KERN_ERR, "I/O resource 0x%x @ 0x%llx busy\n",
		    WIN_SIZE, (u64) pci_resource_start(pdev,1));
	    goto fail_req_bar1;
	}
	pq->bar0 = ioremap_nocache(pci_resource_start(pdev,0), WIN_SIZE);
	if (!pq->bar0)	{
	    module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
		   WIN_SIZE, (u64) pci_resource_start(pdev,0));
	    goto fail_map_bar0;
	}
	pq->bar1 = ioremap_nocache(pci_resource_start(pdev,1), WIN_SIZE);
	if (!pq->bar1)	{
	    module_printk(KERN_ERR, "Can't map 0x%x @ 0x%llx\n",
		   WIN_SIZE, (u64) pci_resource_start(pdev,1));
	    goto fail_map_bar1;
	}

	quasar_sysfs_create (pq);

	pq->hal_ops = &quasarm_hal_ops;	/* QUASARM or QUASARMICRO by default */
	if (QUASAR8(pdev))
		pq->hal_ops = &quasar8_hal_ops;
	if (QUASAR2012(pdev))
		pq->hal_ops = &quasar2012_hal_ops;

	QUASAR_HAL(set_hw_options(pq, noec));

	module_printk(KERN_INFO, "Registering interrupt %d with flags 0x%x\n",
		     pdev->irq, irqflags);
	if (request_irq(pdev->irq, quasar_interrupt_hr,
			irqflags, pq->name, (void *) pq) < 0) {
	    module_printk(KERN_ERR, "Can't register interrupt: %d\n", pdev->irq);
	    goto fail_reg_irq;
	}
    cdev_init(&pq->dev_file, &pmc_e1_fops);
	pq->dev_file.owner = THIS_MODULE;
	pq->dev_file.ops = &pmc_e1_fops;
	if (cdev_add(&pq->dev_file, MKDEV(MAJOR(dev_file_glob),  device_number), 1) )
	{
	    module_printk(KERN_ERR, "Can't register device file for: %d\n", device_number);
	    goto fail_req_file;
	}

	if (ts32) {
		pq->channels = 32;
		module_printk(KERN_WARNING, "Enabling 'ts32 mode', so, will register 32 timeslots per SPAN in DAHDI.\n");
	} else {
		pq->channels = 31;
	}

	device_number++; //for non-hot plug systems, for hot-plug system we probably should change system of numbering...
	spin_lock_init(&pq->dev_lock);
	//mark module as initialised
	pq->registered = 1;

	//DAHDI
	module_printk(KERN_INFO, "Quasar board %d with %d ports found\nHw_fw_rev: 0x%x, slot %d, ec:%d, wd:%d\n",
		      pq->dev_number, 
		      pq->e1_amount + 1, 
		      pq->bar0[BAR0_HW_FW_REV] & 0xffffff, 
		      PCI_SLOT(pdev->devfn)+1, 
		      pq->ec, 
		      pq->wd);
	err = qdahdi_register(pq);
	if(err)
	    module_printk(KERN_ERR, "Quasar dahdi registration error.\n");

	//initialising board registers
	QUASAR_HAL(init_core_regs(pq));

	//init framer registers
	QUASAR_HAL(init_framer_regs(pq));

	//initialising time-slots maping table.
	QUASAR_HAL(init_switch_map(pq));
    
    return 0;

//Clean up all resources, but leave in list.
fail_req_file: free_irq(pdev->irq, (void *) pq);

fail_reg_irq:
        quasar_sysfs_delete (pq);
	iounmap((void *)pq->bar1);
fail_map_bar1: iounmap((void *)pq->bar0);
fail_map_bar0: release_mem_region(pci_resource_start(pdev, 1), WIN_SIZE);//unreg 1 bar;
fail_req_bar1: release_mem_region(pci_resource_start(pdev, 0), WIN_SIZE);//unreg 0 bar;
fail_req_bar0: pq->registered = 0;
return -EBUSY;
}

static void quasar_remove_one (struct pci_dev *pdev)
{
    struct list_head *pos, *n;
	struct quasar *pq=NULL;
	list_for_each_safe(pos, n, &dev_list_head)
	{
		if (list_entry(pos,struct quasar,dev_node)->pci_device == pdev)
		{
			pq = list_entry(pos, struct quasar,dev_node);
		}
	}
	if(pq)
	{
		if(pq->registered)
		{
			qdahdi_unregister(pq); //DAHDI
			free_irq(pdev->irq, (void *) pq);
			cdev_del(&pq->dev_file);
			quasar_sysfs_delete (pq);

			iounmap((void *)pq->bar0);
			release_mem_region (pci_resource_start(pdev, 0), WIN_SIZE);
			iounmap((void *)pq->bar1);
			release_mem_region (pci_resource_start(pdev, 1), WIN_SIZE);

		}
		//unlink from device's list
		list_del(&pq->dev_node);
		kfree(pq);
	}
}

static struct pci_driver quasar_driver = {
    name:		DRV_NAME,
	probe:		quasar_init_one,
	remove:		quasar_remove_one,
	id_table:	quasar_tbl,
};

static int __init quasar_init_module(void)
{
	int result, error;
	module_printk(KERN_INFO, "Quasar board driver, Parabel Ltd. http://parabel.ru, "
		      "Version: %s\n", QUASAR_VERSION);

	if (PMC_E1_MAJOR)
	{
		dev_file_glob = MKDEV(PMC_E1_MAJOR,0);
		result = register_chrdev_region(dev_file_glob, PMC_E18_NR, DRV_NAME);
	}
	else
	{
		result = alloc_chrdev_region(&dev_file_glob, 0, PMC_E18_NR, DRV_NAME);
	}
	if (result < 0)
	{
		module_printk(KERN_ERR, "Can't register major number for driver.\n");
		return -EBUSY;
	}
#ifdef PMC_E1_UDEV
       pmc_e1_class = class_create(THIS_MODULE, DRV_NAME);
	if (IS_ERR(pmc_e1_class))
	{
		module_printk(KERN_ERR, "Can't create class " DRV_NAME ".\n");
		error=-PTR_ERR(pmc_e1_class);
		goto fail_reg_class;
	}
#endif

	result = pci_register_driver(&quasar_driver);
	if(result >= 0)
	{
		return 0;
	}
	else
	{
		module_printk(KERN_ERR, "PCI board registration failed!");
		error=-EBUSY;
	}
	pci_unregister_driver(&quasar_driver);

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
fail_reg_class:
#endif

	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	return error;
}

static void __exit quasar_cleanup_module(void)
{
	pci_unregister_driver (&quasar_driver);
	unregister_chrdev_region(dev_file_glob, PMC_E18_NR);
	module_printk(KERN_INFO, "Quasar board driver unloaded.\n");

#ifdef PMC_E1_UDEV
	class_destroy(pmc_e1_class);
#endif

    return;
}

module_init(quasar_init_module);
module_exit(quasar_cleanup_module);

module_param(cacheline, int, 0400);
MODULE_PARM_DESC(cacheline, "PCI Cache Line Size (in units of DWORDs).");
module_param(cacheline_bridge, int, 0400);
MODULE_PARM_DESC(cacheline_bridge, "PCIe bridge Cache Line Size (DWORDs).");
module_param(latency, int, 0400);
MODULE_PARM_DESC(latency, "PCI BUS latency timer.");
module_param(irqflags, int, 0400);
MODULE_PARM_DESC(irqflags, "Interrupt handler flags (passed to request_irq).");
module_param(pcie_prefetch, int, 0400);
MODULE_PARM_DESC(pcie_prefetch, "Enable PCI express prefetch.");

module_param(noec, int, 0400);
MODULE_PARM_DESC(noec, "Force echo canceller off");
module_param(ts32, int, 0400);
MODULE_PARM_DESC(ts32, "Pass 32 timeslots to DAHDI. Used in unframed mode.");

MODULE_AUTHOR("Khelik Mikhail <mixel.net@gmail.com>, Pavel Selivanov <pavel[AT]parabel.ru>");
MODULE_DESCRIPTION("Quasar interface board driver");
MODULE_LICENSE("GPL");
MODULE_VERSION (QUASAR_VERSION);
